@@ -40,5 +40,3 @@ except PkgConfigError:
             extra_compile_args=_extra_compile_args,
             extra_link_args=_extra_link_args)
 
-if __name__ == '__main__':
-    pass
