@@ -10,6 +10,7 @@ API calls.
 import os
 import binascii
 import warnings
+import logging
 
 try:
     from .epan2_ext import lib as epan_lib
@@ -24,18 +25,24 @@ except ImportError:
         warnings.warn("Epan Library Extensions Not Found.")
         raise EpanLoadError
 
+_logger = logging.getLogger(__name__)
+
 class EpanLoadError(Exception):
     pass
 
 _nstime_empty = epan_ffi.new('nstime_t *');
 @epan_ffi.callback('const nstime_t *(*)(struct packet_provider_data *prov, guint32 frame_num)')
 def _wishpy_get_ts(prov, frame_num):
+
+    if prov[0].ref[0].num == frame_num:
+        return epan_ffi.addressof(prov[0].ref[0], 'abs_ts')
+
+    if prov[0].prev[0].num == frame_num:
+        return epan_ffi.addressof(prov[0].prev[0], 'abs_ts')
+
     return _nstime_empty
 
 
-_wishpy_provider_funcs = epan_ffi.new('struct packet_provider_funcs *',
-        [_wishpy_get_ts, epan_ffi.NULL, epan_ffi.NULL, epan_ffi.NULL])
-
 def _epan_init_v2():
 
     epan_lib.init_process_policies()
@@ -72,52 +79,131 @@ def _epan_init_v3():
 
     return result
 
-def _epan_perform_one_packet_dissection_v2(epan_session, hdr, packet_data, cb_func):
-    raise NotImplemented
+def _epan_perform_one_packet_dissection_v2(wishpy_dissector, frames, hdr, packet_data, cb_func):
+    """Performs dissection of a single packet using v2.6
+    """
 
-def _epan_perform_dissection_v2(wth, wth_file_type, cb_func, count=0):
+    # FIXME : hardcoded
+    wth_file_type = 1
+    total_bytes = 0
+
+    epan_session = wishpy_dissector.epan_session
+    epan_dissect_obj = wishpy_dissector.epan_dissector
+    elapsed_time_ptr = wishpy_dissector.elapsed_time_ptr
+    ref_frame_data_ptr = wishpy_dissector.ref_frame_data_ptr
+    last_frame_data = wishpy_dissector.last_frame_data
+
+    if ref_frame_data_ptr[0] == epan_ffi.NULL:
+        curr_frame_data = wishpy_dissector.first_frame_data
+    else:
+        curr_frame_data = epan_ffi.new('frame_data *')
+
+    # Read stuff from `hdr` for us
+    # TODO: Timestamps
+    packet_len = hdr[0].len
+    packet_capture_len = hdr[0].caplen
+
+    # Allocate enough space to 'copy'
+    count = packet_capture_len
+    if packet_capture_len > 1024:
+        count = 4096
+    if packet_capture_len > 4096:
+        _logger.debug("Big Packet: Length: %d", packet_capture_len)
+        count = 16384
+    if packet_capture_len > 16384:
+        _logger.debug("Huge Packet: Length: %d", packet_capture_len)
+        count = 65536
+    alloc_str = 'guint8[{:d}]'.format(count)
+
+    buf_ptr = epan_ffi.new(alloc_str)
+    # Copy from the data `pakcet_data` to us
+    epan_ffi.memmove(buf_ptr, packet_data, packet_capture_len)
+
+    offset = epan_ffi.new('gint64 *')
+    cum_bytes = epan_ffi.new('guint32 *')
+
+    # Go ahead fill-up the `rec`
+    rec = epan_ffi.new('wtap_rec *')
+    epan_lib.wtap_rec_init(rec)
+    rec[0].rec_header.packet_header.len = packet_len
+    rec[0].rec_header.packet_header.caplen = packet_capture_len
+    rec[0].rec_header.packet_header.pkt_encap = 1 # FIXME: Hard coded
+    rec[0].ts.secs = hdr[0].ts.tv_sec
+    rec[0].ts.nsecs = hdr[0].ts.tv_usec * 1000 # Asumes usec precision FIXME
+    rec[0].presence_flags = epan_lib.WTAP_HAS_TS | epan_lib.WTAP_HAS_CAP_LEN
+
+    epan_lib.frame_data_init(curr_frame_data, frames, rec,
+            offset[0], cum_bytes[0])
+
+    epan_lib.frame_data_set_before_dissect(curr_frame_data,
+            elapsed_time_ptr, ref_frame_data_ptr, last_frame_data)
+
+    # Not sure what this is - Just copied from `tshark` sources
+    epan_lib.prime_epan_dissect_with_postdissector_wanted_hfids(epan_dissect_obj)
+
+    ## Do actual dissection here.
+    # Get buffer and tvbuff first and then run dissector
+    tvb_ptr = epan_lib.tvb_new_real_data(buf_ptr, packet_len, packet_capture_len)
+    epan_lib.epan_dissect_run(epan_dissect_obj, wth_file_type, rec,
+            tvb_ptr, curr_frame_data, epan_ffi.NULL)
+
+    dissected = cb_func(epan_dissect_obj)
+
+    # Get into last data - useful for relative analysis
+    epan_ffi.memmove(last_frame_data,
+            curr_frame_data, epan_ffi.sizeof('frame_data'))
+
+    # Reset the frame data and dissector object
+    epan_lib.frame_data_set_after_dissect(curr_frame_data, cum_bytes)
+    epan_lib.epan_dissect_reset(epan_dissect_obj)
+
+    return dissected
+
+def _epan_perform_dissection_v2(wishpy_dissector, wth, wth_file_type, cb_func, count=0, skip=-1):
     """
     Performs dissection of the file bound to `wth`
     """
-    #empty_packet_provider_funcs = epan_ffi.new('struct packet_provider_funcs *')
-    epan_session = epan_lib.epan_new(epan_ffi.NULL, _wishpy_provider_funcs)
 
-    # TODO: make proper `epan_session` for us
-    epan_dissect_obj = epan_lib.epan_dissect_new(epan_session, True, True)
+    epan_session = wishpy_dissector.epan_session
+    epan_dissect_obj = wishpy_dissector.epan_dissector
+    elapsed_time_ptr = wishpy_dissector.elapsed_time_ptr
+    ref_frame_data_ptr = wishpy_dissector.ref_frame_data_ptr
+    last_frame_data = wishpy_dissector.last_frame_data
 
     offset = epan_ffi.new('gint64 *')
-    frame_data_ref = epan_ffi.new('frame_data **')
-    elapsed_time_ptr = epan_ffi.new('nstime_t *')
 
     err = epan_ffi.new('int *')
     err_str = epan_ffi.new("gchar **")
 
-    frame_data_ref[0] = epan_ffi.NULL
-    frame_data_ptr = epan_ffi.new('frame_data *')
     cum_bytes = epan_ffi.new('guint32 *')
-    processed = 0
+    processed = 1
     total_bytes = 0
     while True:
+
+        if ref_frame_data_ptr[0] == epan_ffi.NULL:
+            curr_frame_data = wishpy_dissector.first_frame_data
+        else:
+            curr_frame_data = epan_ffi.new('frame_data *')
+
         result = epan_lib.wtap_read(wth, err, err_str, offset)
 
         if result == True:
-            processed += 1
+
             rec = epan_lib.wtap_get_rec(wth)
             buf_ptr = epan_lib.wtap_get_buf_ptr(wth)
 
-            epan_rec = epan_ffi.cast('wtap_rec *', rec)
             pkt_len = rec.rec_header.packet_header.len
             pkt_reported_len = rec.rec_header.packet_header.caplen
 
-
-            epan_lib.frame_data_init(frame_data_ptr, processed, epan_rec,
+            epan_lib.frame_data_init(curr_frame_data, processed, rec,
                     offset[0], cum_bytes[0])
 
             total_bytes += pkt_reported_len
             cum_bytes[0] = total_bytes
+
             # FIXME: Look at properly using `frame_data_ref`
-            epan_lib.frame_data_set_before_dissect(frame_data_ptr,
-                    elapsed_time_ptr, frame_data_ref, epan_ffi.NULL)
+            epan_lib.frame_data_set_before_dissect(curr_frame_data,
+                    elapsed_time_ptr, ref_frame_data_ptr, last_frame_data)
 
             # Not sure what this is - Just copied from `tshark` sources
             epan_lib.prime_epan_dissect_with_postdissector_wanted_hfids(
@@ -129,12 +215,16 @@ def _epan_perform_dissection_v2(wth, wth_file_type, cb_func, count=0):
                     pkt_reported_len)
 
             epan_lib.epan_dissect_run(epan_dissect_obj, wth_file_type,
-                    epan_rec, tvb_ptr, frame_data_ptr, epan_ffi.NULL)
+                    rec, tvb_ptr, curr_frame_data, epan_ffi.NULL)
 
             dissected = cb_func(epan_dissect_obj)
 
+            # Get into last data - useful for relative analysis
+            epan_ffi.memmove(last_frame_data,
+                curr_frame_data, epan_ffi.sizeof('frame_data'))
+
             # Reset the frame data and dissector object
-            epan_lib.frame_data_set_after_dissect(frame_data_ptr, cum_bytes)
+            epan_lib.frame_data_set_after_dissect(curr_frame_data, cum_bytes)
             epan_lib.epan_dissect_reset(epan_dissect_obj)
 
             processed += 1
@@ -147,97 +237,104 @@ def _epan_perform_dissection_v2(wth, wth_file_type, cb_func, count=0):
         else:
             break
 
-    epan_lib.epan_dissect_free(epan_dissect_obj)
-    epan_lib.epan_free(epan_session)
-
-def _epan_perform_one_packet_dissection_v3(hdr, packet_data, cb_func):
+def _epan_perform_one_packet_dissection_v3(wishpy_dissector, frames, hdr, packet_data, cb_func):
     """Performs a single packet dissection.
     """
 
     # FIXME : hardcoded
     wth_file_type = 1
     total_bytes = 0
-    # FIXME: epan_session and epan_dissect_obj to be set by caller? So that
-    # We don't free on every data.
-    epan_session = epan_lib.epan_new(epan_ffi.NULL, _wishpy_provider_funcs)
 
-    epan_dissect_obj = epan_lib.epan_dissect_new(epan_session, True, True)
-
-    offset = epan_ffi.new('gint64 *')
-    frame_data_ref = epan_ffi.new('frame_data **')
-    elapsed_time_ptr = epan_ffi.new('nstime_t *')
-
-    buf = epan_ffi.new('Buffer *')
-    epan_lib.ws_buffer_init(buf, 1514) # FIXME : Should do with proper length
-    start_ptr = buf[0].data + buf[0].start
-
-    rec = epan_ffi.new('wtap_rec *')
-    epan_lib.wtap_rec_init(rec)
-
-    err = epan_ffi.new('int *')
-    err_str = epan_ffi.new("gchar **")
-
-    frame_data_ref[0] = epan_ffi.NULL
-    frame_data_ptr = epan_ffi.new('frame_data *')
-    cum_bytes = epan_ffi.new('guint32 *')
+    epan_session = wishpy_dissector.epan_session
+    epan_dissect_obj = wishpy_dissector.epan_dissector
+    elapsed_time_ptr = wishpy_dissector.elapsed_time_ptr
+    ref_frame_data_ptr = wishpy_dissector.ref_frame_data_ptr
+    last_frame_data = wishpy_dissector.last_frame_data
 
+    if ref_frame_data_ptr[0] == epan_ffi.NULL:
+        curr_frame_data = wishpy_dissector.first_frame_data
+    else:
+        curr_frame_data = epan_ffi.new('frame_data *')
+    # Read stuff from `hdr` for us
     packet_len = hdr[0].len
     packet_capture_len = hdr[0].caplen
 
+    # Allocate enough space to 'copy'
+    count = packet_capture_len
+    if packet_capture_len > 1024:
+        count = 4096
+    if packet_capture_len > 4096:
+        _logger.debug("Big Packet: Length: %d", packet_capture_len)
+        count = 16384
+    if packet_capture_len > 16384:
+        _logger.debug("Huge Packet: Length: %d", packet_capture_len)
+        count = 65536
+    count = packet_capture_len
+
+    # Copy from the data `packet_data` to us
+    # We've to oblige to `wireshark` way of doing it hence the heavy lifting
+    buf = epan_ffi.new('Buffer *')
+    epan_lib.ws_buffer_init(buf, count)
+    start_ptr = buf[0].data + buf[0].start
+    epan_ffi.memmove(start_ptr, packet_data, packet_capture_len)
+
+
+    rec = epan_ffi.new('wtap_rec *')
+    epan_lib.wtap_rec_init(rec)
     rec[0].rec_header.packet_header.len = packet_len
     rec[0].rec_header.packet_header.caplen = packet_capture_len
     rec[0].rec_header.packet_header.pkt_encap = 1 # FIXME: Hard coded
+    rec[0].ts.secs = hdr[0].ts.tv_sec
+    rec[0].ts.nsecs = hdr[0].ts.tv_usec * 1000 # Asumes usec precision FIXME
+    rec[0].presence_flags = epan_lib.WTAP_HAS_TS | epan_lib.WTAP_HAS_CAP_LEN
 
-    epan_lib.frame_data_init(frame_data_ptr, 0, rec, offset[0], cum_bytes[0])
+    offset = epan_ffi.new('gint64 *')
+    cum_bytes = epan_ffi.new('guint32 *')
+    epan_lib.frame_data_init(curr_frame_data, frames, rec, offset[0], cum_bytes[0])
 
     total_bytes += packet_capture_len
     cum_bytes[0] = total_bytes
 
-    # Copy from the data to us
-    epan_ffi.memmove(start_ptr, packet_data, packet_capture_len)
-
-    # FIXME: Look at properly using `frame_data_ref`
-    epan_lib.frame_data_set_before_dissect(frame_data_ptr,
-            elapsed_time_ptr, frame_data_ref, epan_ffi.NULL)
+    epan_lib.frame_data_set_before_dissect(curr_frame_data,
+            elapsed_time_ptr, ref_frame_data_ptr, last_frame_data)
 
     # Not sure what this is - Just copied from `tshark` sources
     epan_lib.prime_epan_dissect_with_postdissector_wanted_hfids(
             epan_dissect_obj)
 
-    ## Do actual dissection here.
+    ## Ready to do `dissection` here.
     # Get buffer and tvbuff first and then run dissector
     tvb_ptr = epan_lib.tvb_new_real_data(buf[0].data, packet_len,
             packet_capture_len)
 
     epan_lib.epan_dissect_run(epan_dissect_obj, wth_file_type,
-            rec, tvb_ptr, frame_data_ptr, epan_ffi.NULL)
+            rec, tvb_ptr, curr_frame_data, epan_ffi.NULL)
 
     dissected = cb_func(epan_dissect_obj)
 
+    # Get into last data - useful for relative analysis
+    epan_ffi.memmove(last_frame_data,
+            curr_frame_data, epan_ffi.sizeof('frame_data'))
     # Reset the frame data and dissector object
-    epan_lib.frame_data_set_after_dissect(frame_data_ptr, cum_bytes)
+    epan_lib.frame_data_set_after_dissect(curr_frame_data, cum_bytes)
     epan_lib.epan_dissect_reset(epan_dissect_obj)
 
-    epan_lib.epan_dissect_free(epan_dissect_obj)
-    epan_lib.epan_free(epan_session)
-
     return dissected
 
 
-def _epan_perform_dissection_v3(wth, wth_file_type, cb_func, count=0, skip=-1):
+def _epan_perform_dissection_v3(wishpy_dissector, wth, wth_file_type, cb_func, count=0, skip=-1):
     """
     Performs dissection of the file bound to `wth`. If Non-zero positive count
     is specified, performs dissection of up to `count` packets
     """
-    #empty_packet_provider_funcs = epan_ffi.new('struct packet_provider_funcs *')
-    epan_session = epan_lib.epan_new(epan_ffi.NULL, _wishpy_provider_funcs)
 
-    # TODO: make proper `epan_session` for us
-    epan_dissect_obj = epan_lib.epan_dissect_new(epan_session, True, True)
+    epan_session = wishpy_dissector.epan_session
+    epan_dissect_obj = wishpy_dissector.epan_dissector
+    elapsed_time_ptr = wishpy_dissector.elapsed_time_ptr
+    ref_frame_data_ptr = wishpy_dissector.ref_frame_data_ptr
+    last_frame_data = wishpy_dissector.last_frame_data
 
     offset = epan_ffi.new('gint64 *')
-    frame_data_ref = epan_ffi.new('frame_data **')
-    elapsed_time_ptr = epan_ffi.new('nstime_t *')
 
     buf = epan_ffi.new('Buffer *')
     epan_lib.ws_buffer_init(buf, 1514) # FIXME : Should do with proper length
@@ -246,13 +343,17 @@ def _epan_perform_dissection_v3(wth, wth_file_type, cb_func, count=0, skip=-1):
     err = epan_ffi.new('int *')
     err_str = epan_ffi.new("gchar **")
 
-    frame_data_ref[0] = epan_ffi.NULL
-    frame_data_ptr = epan_ffi.new('frame_data *')
     cum_bytes = epan_ffi.new('guint32 *')
-    processed = 0
+    processed = 1
     total_bytes = 0
     skipped = 0
     while True:
+
+        if ref_frame_data_ptr[0] == epan_ffi.NULL:
+            curr_frame_data = wishpy_dissector.first_frame_data
+        else:
+            curr_frame_data = epan_ffi.new('frame_data *')
+
         result = epan_lib.wtap_read(wth, rec, buf, err, err_str, offset)
 
         if result == True:
@@ -260,15 +361,14 @@ def _epan_perform_dissection_v3(wth, wth_file_type, cb_func, count=0, skip=-1):
             pkt_len = rec[0].rec_header.packet_header.len
             pkt_reported_len = rec[0].rec_header.packet_header.caplen
 
-            epan_lib.frame_data_init(frame_data_ptr, processed, rec,
+            epan_lib.frame_data_init(curr_frame_data, processed, rec,
                     offset[0], cum_bytes[0])
 
             total_bytes += pkt_reported_len
             cum_bytes[0] = total_bytes
 
-            # FIXME: Look at properly using `frame_data_ref`
-            epan_lib.frame_data_set_before_dissect(frame_data_ptr,
-                    elapsed_time_ptr, frame_data_ref, epan_ffi.NULL)
+            epan_lib.frame_data_set_before_dissect(curr_frame_data,
+                    elapsed_time_ptr, ref_frame_data_ptr, last_frame_data)
 
             # Not sure what this is - Just copied from `tshark` sources
             epan_lib.prime_epan_dissect_with_postdissector_wanted_hfids(
@@ -280,7 +380,7 @@ def _epan_perform_dissection_v3(wth, wth_file_type, cb_func, count=0, skip=-1):
                     pkt_reported_len)
 
             epan_lib.epan_dissect_run(epan_dissect_obj, wth_file_type,
-                    rec, tvb_ptr, frame_data_ptr, epan_ffi.NULL)
+                    rec, tvb_ptr, curr_frame_data, epan_ffi.NULL)
 
             if skipped < skip:
                 skipped += 1
@@ -288,8 +388,12 @@ def _epan_perform_dissection_v3(wth, wth_file_type, cb_func, count=0, skip=-1):
 
             dissected = cb_func(epan_dissect_obj)
 
+            # Get into last data - useful for relative analysis
+            epan_ffi.memmove(last_frame_data,
+                    curr_frame_data, epan_ffi.sizeof('frame_data'))
+
             # Reset the frame data and dissector object
-            epan_lib.frame_data_set_after_dissect(frame_data_ptr, cum_bytes)
+            epan_lib.frame_data_set_after_dissect(curr_frame_data, cum_bytes)
             epan_lib.epan_dissect_reset(epan_dissect_obj)
 
             processed += 1
@@ -302,9 +406,6 @@ def _epan_perform_dissection_v3(wth, wth_file_type, cb_func, count=0, skip=-1):
         else:
             break
 
-    epan_lib.epan_dissect_free(epan_dissect_obj)
-    epan_lib.epan_free(epan_session)
-
     return processed
 
 def wtap_open_file_offline(filepath):
@@ -314,6 +415,7 @@ def wtap_open_file_offline(filepath):
     """
 
     if not os.path.exists(filepath):
+        _logger.error("File not found: %s", filepath)
         return None, None
 
     # Open a wtap file
@@ -323,6 +425,8 @@ def wtap_open_file_offline(filepath):
     open_type = epan_lib.WTAP_TYPE_AUTO
     wth = epan_lib.wtap_open_offline(filename.encode(), open_type, err, err_str, False)
     if wth is epan_ffi.NULL:
+        err_display_str = epan_ffi.string(err_str)
+        _logger.error("Error creating wiretap handle: %s", err_display_str)
         return None, None
 
     wtap_file_type = epan_lib.wtap_file_type_subtype(wth)
@@ -331,6 +435,35 @@ def wtap_open_file_offline(filepath):
 wtap_close = epan_lib.wtap_close
 epan_cleanup = epan_lib.epan_cleanup
 
+_wishpy_provider_funcs = epan_ffi.new('struct packet_provider_funcs *',
+        [_wishpy_get_ts, epan_ffi.NULL, epan_ffi.NULL, epan_ffi.NULL])
+
+
+def epan_new_session(provider=epan_ffi.NULL):
+    """Get new `epan_session`
+
+    Calls the wrapped `epan_new` and returns the handler.
+    """
+    return epan_lib.epan_new(provider, _wishpy_provider_funcs)
+
+def epan_free_session(session):
+    """Calls `epan_free` on the given `session` object.
+    """
+    epan_lib.epan_free(session)
+
+def epan_new_dissector(session):
+    """Get new `epan_dissect_t`
+
+    Calls the wrapped `epan_dissect_new` and returns the handler.
+    Params:
+        session: `epan_session` object returned by `epan_new_session`(ish).
+    """
+    return epan_lib.epan_dissect_new(session, True, True)
+
+def epan_free_dissector(epan_dissector_obj):
+    """Calls internal `epan_dissect_free` on the `epan_dissector_obj`
+    """
+    epan_lib.epan_dissect_free(epan_dissector_obj)
 
 if _epan_version == (2,6):
     epan_lib_init = _epan_init_v2
@@ -343,6 +476,8 @@ else:
 
 def perform_epan_wtap_init():
 
+    _logger.info("Found Wireshark Epan Version %d.%d", _epan_version[0], _epan_version[1])
+
     # FIXME: True / False to be decided
     epan_lib.wtap_init(False)
 
